@@ -7,7 +7,11 @@
 #include "irrMath.h"
 
 #if defined(_IRR_COMPILE_WITH_SDL_DEVICE_)
-#include <SDL_endian.h>
+#ifdef _IRR_USE_SDL3_
+	#include <SDL3/SDL_endian.h>
+#else
+	#include <SDL_endian.h>
+#endif
 #define bswap_16(X) SDL_Swap16(X)
 #define bswap_32(X) SDL_Swap32(X)
 #define bswap_64(X) SDL_Swap64(X)
