@@ -8,7 +8,7 @@
 #include "inputhandler.h"
 #include "gui/mainmenumanager.h"
 #include "gui/touchcontrols.h"
-#include "hud.h"
+#include "hud_element.h"
 #include "log_internal.h"
 #include "client/renderingengine.h"
 
@@ -26,7 +26,7 @@ void MyEventReceiver::reloadKeybindings()
 	keybindings[KeyType::DIG] = getKeySetting("keymap_dig");
 	keybindings[KeyType::PLACE] = getKeySetting("keymap_place");
 
-	keybindings[KeyType::ESC] = EscapeKey;
+	keybindings[KeyType::ESC] = std::vector<KeyPress>{EscapeKey};
 
 	keybindings[KeyType::AUTOFORWARD] = getKeySetting("keymap_autoforward");
 
@@ -76,7 +76,9 @@ void MyEventReceiver::reloadKeybindings()
 	// First clear all keys, then re-add the ones we listen for
 	keysListenedFor.clear();
 	for (int i = 0; i < KeyType::INTERNAL_ENUM_COUNT; i++) {
-		listenForKey(keybindings[i], static_cast<GameKeyType>(i));
+		for (auto key: keybindings[i]) {
+			listenForKey(key, static_cast<GameKeyType>(i));
+		}
 	}
 }
 
@@ -85,13 +87,11 @@ bool MyEventReceiver::setKeyDown(KeyPress keyCode, bool is_down)
 	if (keysListenedFor.find(keyCode) == keysListenedFor.end()) // ignore irrelevant key input
 		return false;
 	auto action = keysListenedFor[keyCode];
-	if (is_down) {
+	if (is_down)
 		physicalKeyDown.insert(keyCode);
-		setKeyDown(action, true);
-	} else {
+	else
 		physicalKeyDown.erase(keyCode);
-		setKeyDown(action, false);
-	}
+	setKeyDown(action, checkKeyDown(action));
 	return true;
 }
 
@@ -109,9 +109,18 @@ void MyEventReceiver::setKeyDown(GameKeyType action, bool is_down)
 	}
 }
 
+bool MyEventReceiver::checkKeyDown(GameKeyType action) const
+{
+	for (const auto &key : keybindings[action]) {
+		if (physicalKeyDown.find(key) != physicalKeyDown.end())
+			return true;
+	}
+	return false;
+}
+
 bool MyEventReceiver::OnEvent(const SEvent &event)
 {
-	if (event.EventType == irr::EET_LOG_TEXT_EVENT) {
+	if (event.EventType == EET_LOG_TEXT_EVENT) {
 		static const LogLevel irr_loglev_conv[] = {
 			LL_VERBOSE, // ELL_DEBUG
 			LL_INFO,    // ELL_INFORMATION
@@ -137,21 +146,28 @@ bool MyEventReceiver::OnEvent(const SEvent &event)
 	// This is separate from other keyboard handling so that it also works in menus.
 	if (event.EventType == EET_KEY_INPUT_EVENT) {
 		KeyPress keyCode(event.KeyInput);
-		if (keyCode == getKeySetting("keymap_fullscreen")) {
+
+		if (keySettingHasMatch("keymap_fullscreen", keyCode)) {
 			if (event.KeyInput.PressedDown && !fullscreen_is_down) {
 				IrrlichtDevice *device = RenderingEngine::get_raw_device();
 
 				bool new_fullscreen = !device->isFullscreen();
-				// Only update the setting if toggling succeeds - it always fails
-				// if Minetest was built without SDL.
+				// Only update the setting if toggling succeeds
 				if (device->setFullscreen(new_fullscreen)) {
 					g_settings->setBool("fullscreen", new_fullscreen);
 				}
 			}
 			fullscreen_is_down = event.KeyInput.PressedDown;
 			return true;
-		} else if (keyCode == EscapeKey &&
-				event.KeyInput.PressedDown && event.KeyInput.Shift) {
+
+		} else if (keySettingHasMatch("keymap_close_world", keyCode)) {
+			close_world_down = event.KeyInput.PressedDown;
+
+		} else if (keyCode == EscapeKey) {
+			esc_down = event.KeyInput.PressedDown;
+		}
+
+		if (esc_down && close_world_down) {
 			g_gamecallback->disconnect();
 			return true;
 		}
@@ -170,18 +186,18 @@ bool MyEventReceiver::OnEvent(const SEvent &event)
 	}
 
 	// Remember whether each key is down or up
-	if (event.EventType == irr::EET_KEY_INPUT_EVENT) {
+	if (event.EventType == EET_KEY_INPUT_EVENT) {
 		KeyPress keyCode(event.KeyInput);
 		if (setKeyDown(keyCode, event.KeyInput.PressedDown))
 			return true;
-	} else if (g_touchcontrols && event.EventType == irr::EET_TOUCH_INPUT_EVENT) {
+	} else if (g_touchcontrols && event.EventType == EET_TOUCH_INPUT_EVENT) {
 		// In case of touchcontrols, we have to handle different events
 		g_touchcontrols->translateEvent(event);
 		return true;
-	} else if (event.EventType == irr::EET_JOYSTICK_INPUT_EVENT) {
+	} else if (event.EventType == EET_JOYSTICK_INPUT_EVENT) {
 		// joystick may be nullptr if game is launched with '--random-input' parameter
 		return joystick && joystick->handleEvent(event.JoystickEvent);
-	} else if (event.EventType == irr::EET_MOUSE_INPUT_EVENT) {
+	} else if (event.EventType == EET_MOUSE_INPUT_EVENT) {
 		// Handle mouse events
 		switch (event.MouseInput.Event) {
 		case EMIE_LMOUSE_PRESSED_DOWN:
