@@ -14,15 +14,15 @@
 #include "util/numeric.h"
 #include "light.h"
 #include "localplayer.h"
-#include "environment.h"
 #include "clientmap.h"
 #include "mapnode.h"
+#include "node_visuals.h"
 #include "nodedef.h"
 #include "client.h"
 #include "settings.h"
 #include "profiler.h"
 
-#include "SMeshBuffer.h"
+#include "CMeshBuffer.h"
 
 using BlendMode = ParticleParamTypes::BlendMode;
 
@@ -31,6 +31,25 @@ ClientParticleTexture::ClientParticleTexture(const ServerParticleTexture& p, ITe
 	tex = p;
 	// note: getTextureForMesh not needed here because we don't use texture filtering
 	ref = tsrc->getTexture(p.string);
+
+	// Try to show another texture to indicate a code issue.
+	if (!ref)
+		ref = tsrc->getTexture("no_texture.png");
+}
+
+static video::ITexture *extractTexture(const TileDef &def, const TileLayer &layer,
+	ITextureSource *tsrc)
+{
+	// If animated take first frame from tile layer (so we don't have to handle
+	// that manually), otherwise look up by name.
+	if (!layer.empty() && (layer.material_flags & MATERIAL_FLAG_ANIMATION)) {
+		auto *ret = (*layer.frames)[0].texture;
+		assert(ret->getType() == video::ETT_2D);
+		return ret;
+	}
+	if (!def.name.empty())
+		return tsrc->getTexture(def.name);
+	return nullptr;
 }
 
 /*
@@ -193,7 +212,7 @@ void Particle::updateVertices(ClientEnvironment *env, video::SColor color)
 	video::S3DVertex *vertices = m_buffer->getVertices(m_index);
 
 	if (m_texture.tex != nullptr)
-		scale = m_texture.tex -> scale.blend(m_time / (m_expiration+0.1));
+		scale = m_texture.tex -> scale.blend(m_time / (m_expiration+0.1f));
 	else
 		scale = v2f(1.f, 1.f);
 
@@ -203,7 +222,7 @@ void Particle::updateVertices(ClientEnvironment *env, video::SColor color)
 		v2u32 framesize;
 		texcoord = m_p.animation.getTextureCoords(texsize, m_animation_frame);
 		m_p.animation.determineParams(texsize, NULL, NULL, &framesize);
-		framesize_f = v2f(framesize.X / (float) texsize.X, framesize.Y / (float) texsize.Y);
+		framesize_f = v2f::from(framesize) / v2f::from(texsize);
 
 		tx0 = m_texpos.X + texcoord.X;
 		tx1 = m_texpos.X + texcoord.X + framesize_f.X * m_texsize.X;
@@ -447,22 +466,23 @@ void ParticleSpawner::spawnParticle(ClientEnvironment *env, float radius,
 	video::SColor color(0xFFFFFFFF);
 
 	if (p.node.getContent() != CONTENT_IGNORE) {
-		const ContentFeatures &f =
-			m_particlemanager->m_env->getGameDef()->ndef()->get(p.node);
-		if (!ParticleManager::getNodeParticleParams(p.node, f, pp, &texture.ref,
-				texpos, texsize, &color, p.node_tile))
+		if (!ParticleManager::getNodeParticleParams(env->getGameDef(), p.node,
+				pp, &texture.ref, texpos, texsize, &color, p.node_tile))
 			return;
 	} else {
 		if (m_texpool.size() == 0)
 			return;
-		texture = ClientParticleTexRef(m_texpool[m_texpool.size() == 1 ? 0
-				: myrand_range(0, m_texpool.size()-1)]);
+		texture = ClientParticleTexRef(m_texpool[myrand_range(0, m_texpool.size() - 1)]);
 		texpos = v2f(0.0f, 0.0f);
 		texsize = v2f(1.0f, 1.0f);
 		if (texture.tex->animated)
 			pp.animation = texture.tex->animation;
 	}
 
+	// Same guard as in `CE_SPAWN_PARTICLE`
+	if (!texture.ref)
+		return;
+
 	// synchronize animation length with particle life if desired
 	if (pp.animation.type != TAT_NONE) {
 		// FIXME: this should be moved into a TileAnimationParams class method
@@ -786,7 +806,8 @@ void ParticleManager::handleParticleEvent(ClientEvent *event, Client *client,
 
 			const ParticleSpawnerParameters &p = *event->add_particlespawner.p;
 
-			// texture pool
+			// There can be multiple textures, e.g. for time-based animations
+			// Look up all required textures in `ITextureSource` to retrieve an `ITexture`.
 			std::vector<ClientParticleTexture> texpool;
 			if (!p.texpool.empty()) {
 				size_t txpsz = p.texpool.size();
@@ -822,9 +843,8 @@ void ParticleManager::handleParticleEvent(ClientEvent *event, Client *client,
 			f32 oldsize = p.size;
 
 			if (p.node.getContent() != CONTENT_IGNORE) {
-				const ContentFeatures &f = m_env->getGameDef()->ndef()->get(p.node);
-				getNodeParticleParams(p.node, f, p, &texture.ref, texpos,
-						texsize, &color, p.node_tile);
+				getNodeParticleParams(m_env->getGameDef(), p.node, p,
+						&texture.ref, texpos, texsize, &color, p.node_tile);
 			} else {
 				/* with no particlespawner to own the texture, we need
 				 * to save it on the heap. it will be freed when the
@@ -853,28 +873,29 @@ void ParticleManager::handleParticleEvent(ClientEvent *event, Client *client,
 	}
 }
 
-bool ParticleManager::getNodeParticleParams(const MapNode &n,
-	const ContentFeatures &f, ParticleParameters &p, video::ITexture **texture,
+bool ParticleManager::getNodeParticleParams(Client *client, const MapNode &n,
+	ParticleParameters &p, video::ITexture **texture,
 	v2f &texpos, v2f &texsize, video::SColor *color, u8 tilenum)
 {
+	const ContentFeatures &f = client->ndef()->get(n);
+
 	// No particles for "airlike" nodes
 	if (f.drawtype == NDT_AIRLIKE)
 		return false;
 
 	// Texture
+	// Note: we ignore the overlay here, oh well
 	u8 texid;
 	if (tilenum > 0 && tilenum <= 6)
 		texid = tilenum - 1;
 	else
 		texid = myrand_range(0,5);
-	const TileLayer &tile = f.tiles[texid].layers[0];
-	p.animation.type = TAT_NONE;
 
-	// Only use first frame of animated texture
-	if (tile.material_flags & MATERIAL_FLAG_ANIMATION)
-		*texture = (*tile.frames)[0].texture;
-	else
-		*texture = tile.texture;
+	const TileLayer &tile = f.visuals->tiles[texid].layers[0];
+	*texture = extractTexture(f.tiledef[texid], tile, client->tsrc());
+	p.texture.blendmode = f.alpha == ALPHAMODE_BLEND
+			? BlendMode::alpha : BlendMode::clip;
+	p.animation.type = TAT_NONE;
 
 	float size = (myrand_range(0,8)) / 64.0f;
 	p.size = BS * size;
@@ -887,7 +908,7 @@ bool ParticleManager::getNodeParticleParams(const MapNode &n,
 	if (tile.has_color)
 		*color = tile.color;
 	else
-		n.getColor(f, color);
+		f.visuals->getColor(n.param2, color);
 
 	return true;
 }
@@ -895,35 +916,26 @@ bool ParticleManager::getNodeParticleParams(const MapNode &n,
 // The final burst of particles when a node is finally dug, *not* particles
 // spawned during the digging of a node.
 
-void ParticleManager::addDiggingParticles(IGameDef *gamedef,
-	LocalPlayer *player, v3s16 pos, const MapNode &n, const ContentFeatures &f)
+void ParticleManager::addDiggingParticles(LocalPlayer *player, v3s16 pos, const MapNode &n)
 {
-	// No particles for "airlike" nodes
-	if (f.drawtype == NDT_AIRLIKE)
-		return;
-
 	for (u16 j = 0; j < 16; j++) {
-		addNodeParticle(gamedef, player, pos, n, f);
+		addNodeParticle(player, pos, n);
 	}
 }
 
 // During the digging of a node particles are spawned individually by this
 // function, called from Game::handleDigging() in game.cpp.
 
-void ParticleManager::addNodeParticle(IGameDef *gamedef,
-	LocalPlayer *player, v3s16 pos, const MapNode &n, const ContentFeatures &f)
+void ParticleManager::addNodeParticle(LocalPlayer *player, v3s16 pos, const MapNode &n)
 {
 	ParticleParameters p;
 	video::ITexture *ref = nullptr;
 	v2f texpos, texsize;
 	video::SColor color;
 
-	if (!getNodeParticleParams(n, f, p, &ref, texpos, texsize, &color))
+	if (!getNodeParticleParams(m_env->getGameDef(), n, p, &ref, texpos, texsize, &color))
 		return;
 
-	p.texture.blendmode = f.alpha == ALPHAMODE_BLEND
-			? BlendMode::alpha : BlendMode::clip;
-
 	p.expirationtime = myrand_range(0, 100) / 100.0f;
 
 	// Physics
